(* non-embedded, simple static dispatch *) 



use "../lib/smlprelude";




(* root objects -------------------------------------------------------	*)

open Closed;   (* closed datatype, % : closed *)

signature ROOT =
sig
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : unit -> 'a -> 'a subself
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a;
type self = closed subself;

(* packing *)
fun unpack ob = (I,ob);
fun pack () ob = ob;

(* new *)
fun new () = pack () %;
end;



(* points -------------------------------------------------------------	*)

signature POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Root.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : int -> 'a -> 'a subself
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val dist : 'a subself -> 'b subself -> real
    val new : int -> self
end;

structure Point :> POINT =
struct

(* types *)
type 'a ext = int * 'a;
type 'a subself = 'a ext Root.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode f (Root.unpack p);
fun pack n ext = Root.pack () (n, ext);

(* messages *)
fun pos p = let val (_,(x,_)) = Root.unpack p in x end;
fun slide p n = repack (fn x => x+n) p;
fun eq p q = pos p = pos q;
fun dist p q = real (pos q - pos p);

(* new *)
fun new n = pack n %;
end;



(* colored points  ----------------------------------------------------	*)

signature CPOINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself;
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : color * int -> 'a -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
    val new : int * color -> self
end;

structure Cpoint :> CPOINT = 
struct

(* types *)
type 'a ext = color * 'a;
type 'a subself = 'a ext Point.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode f (Point.unpack p);
fun pack (c,n) ext = Point.pack n (c, ext);

(* messages *)
fun col p = let val (_,(c,_)) = Point.unpack p in c end;
fun paint p c = repack (fn _ => c) p;
fun eq p q = col p = col q andalso Point.eq p q;

(* new *)
fun new (n,c) : self = pack (c,n) %;
end;



(* 2d points ---------------------------------------------------------- *)
 
signature D2POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : int * int -> 'a -> 'a subself
    val xpos : 'a subself -> int
    val ypos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val dist : 'a subself -> 'b subself -> real
    val new : int * int -> self
end;

structure D2point :> D2POINT =
struct

(* types *)
type 'a ext = int * 'a;
type 'a subself = 'a ext Point.subself;
type self = closed subself;
 
(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode f (Point.unpack p);
fun pack (y,x) ext = Point.pack x (y, ext);

(* messages *)
val xpos = Point.pos;
fun ypos p = let val (_,(y,_)) = Point.unpack p in y end;
fun slide p n = repack (fn y => y+n) (Point.slide p n);
fun eq p q = Point.eq p q andalso ypos p = ypos q;
fun dist p q = sqrt (Real.+ (sq (real (ypos q - ypos p)), sq (Point.dist p q)));

(* new *)
fun new (x,y) : self = pack (y,x) %;
end; 



(* lines --------------------------------------------------------- *)

signature MINIPOINT =
sig

    type 'a subself
    type self = closed subself
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val dist : 'a subself -> 'b subself -> real
end;

signature LINE =
sig
    structure Param : MINIPOINT
    type ('a,'b) ext
    type ('a,'b) subself = ('a,'b) ext Root.subself
    type 'a self = ('a,closed) subself
    val unpack : ('a,'b) subself -> ('b -> ('a,'b) subself) * 'b
    val pack   : ('a Param.subself * 'a Param.subself) -> 'b -> ('a,'b) subself
    val from   : ('a,'b) subself -> 'a Param.subself
    val to     : ('a,'b) subself -> 'a Param.subself
    val shift  : ('a,'b) subself -> int -> ('a,'b) subself
    val eq     : ('a,'b) subself -> ('c,'d) subself -> bool
    val length : ('a,'b) subself -> real
    val new    : 'a Param.subself * 'a Param.subself -> 'a self
end;

functor Line (structure P : MINIPOINT) :> LINE where Param = P =
struct

structure Param = P;

(* types *)
type ('a,'b) ext = ('a Param.subself * 'a Param.subself) * 'b;
type ('a,'b) subself = ('a,'b) ext Root.subself;
type 'a self = ('a,closed) subself;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode f (Root.unpack p);
fun pack ft ext = Root.pack () (ft, ext);

(* messages *)
fun from l = let val (_,((f,_),_)) = Root.unpack l in f end;
fun to l = let val (_,((_,t),_)) = Root.unpack l in t end;
fun shift l n =	repack (fn (f,t) => (Param.slide f n,Param.slide t n)) l;
fun eq p q = Param.eq (from p) (from q) andalso Param.eq (to p) (to q);
fun length l = Param.dist (from l) (to l);

(* new *)
fun new (f,t) = pack (f,t) %;
end;


structure PointLine =
		Line (structure P = Point);
structure CpointLine =
		Line (structure P = struct open Point; open Cpoint; end);
structure D2pointLine =
		Line (structure P = struct open Point; open D2point; end);


