(* self-application, simple static dispatch *)


use "../lib/smlprelude";


(* root objects -------------------------------------------------------	*)

open Closed;   (* closed datatype, % : closed *)

signature ROOT =
sig
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : unit -> 'a -> 'a subself
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a
type self = closed subself

(* packing *)
fun unpack ob = (I,ob);
fun pack () ob = ob;

(* new *)
fun new () = pack () %;
end;



(* points -------------------------------------------------------------	*)

signature POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Root.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : int -> 'a -> 'a subself
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'a subself -> real
    val eq : 'a subself -> 'a subself -> bool
    val new : int -> self
end;

structure Point :> POINT =
struct

(* types *)
datatype 'a state = S of
		('a subself -> int) *			(* pos	 *)
		('a subself -> int -> 'a subself) * 	(* slide *)
		('a subself -> 'a subself -> real) *	(* dist  *)
		('a subself -> 'a subself -> bool) 	(* eq    *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Root.subself;
type self = closed subself;

(* messages *)
fun pos p =     let val (_,(S (n,_,_,_),_)) = Root.unpack p in n p end;
fun slide p n = let val (_,(S (_,s,_,_),_)) = Root.unpack p in s p n end;
fun dist p q =  let val (_,(S (_,_,d,_),_)) = Root.unpack p in d p q end;
fun eq p q =    let val (_,(S (_,_,_,e),_)) = Root.unpack p in e p q end;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Root.unpack p);
fun pack n ext =
	let fun peq _ p q = pos p = pos q
	    fun methods n =
		let fun mpos p = n
		    fun mslide p n' = repack (fn _ => methods (n + n')) p
		    fun mdist p q = real (pos q - pos p)
		    fun meq p q = pos p = pos q
		in (mpos,mslide,mdist,meq)
		end
	in Root.pack () (S (methods n),ext)
	end;

(* new *)
fun new n = pack n %;
end;



(* colored points  ----------------------------------------------------	*)

signature CPOINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself;
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : color * int -> 'a -> 'a subself
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
    val eq : 'a subself -> 'a subself -> bool
    val new : int * color -> self
end;

structure Cpoint :> CPOINT = 
struct

(* types *)
datatype 'a state = S of
		('a subself -> color) *			(* col	 *)
		('a subself -> color -> 'a subself) *	(* paint *)
		('a subself -> 'a subself -> bool) 	(* eq    *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* messages *)
fun col p =     let val (_,(S (c,_,_),_)) = Point.unpack p in c p end;
fun paint p c = let val (_,(S (_,t,_),_)) = Point.unpack p in t p c end;
fun eq p q =    let val (_,(S (_,_,e),_)) = Point.unpack p in e p q end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Point.unpack p);
fun pack (c,n) ext =
	let fun methods c =
		let fun mcol cp = c
		    fun mpaint cp c' = repack (fn _ => methods c') cp
		    fun meq p q = Point.eq p q andalso (col p = col q)
		in (mcol,mpaint,meq)
		end
	in Point.pack n (S (methods c),ext)
	end;

(* new *)
fun new (n,c) = pack (c,n) %;
end;



(* d2point ----------------------------------------------------	*)

signature D2POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : int * int -> 'a -> 'a subself
    val xpos : 'a Point.subself -> int
    val ypos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'a subself -> real
    val eq : 'a subself -> 'a subself -> bool
    val new : int * int -> self
end;

structure D2point :> D2POINT =
struct

(* types *)
datatype 'a state = S of
		('a subself -> int) *			(* ypos  *)
		('a subself -> int -> 'a subself) *	(* slide *)
		('a subself -> 'a subself -> real) *	(* dist  *)
		('a subself -> 'a subself -> bool)	(* eq    *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* messages *)
val xpos = Point.pos; (* alias *)
fun ypos p =    let val (_,(S (y,_,_,_),_)) = Point.unpack p in y p end;
fun slide p n = let val (_,(S (_,s,_,_),_)) = Point.unpack p in s p n end;
fun dist p q =  let val (_,(S (_,_,d,_),_)) = Point.unpack p in d p q end;
fun eq p q =    let val (_,(S (_,_,_,e),_)) = Point.unpack p in e p q end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Point.unpack p);
fun pack (x,y) ext =
	let fun methods y =
		let fun mypos p = y
		    fun mslide p n =
			repack (fn _ => methods (y+n)) (Point.slide p n)
		    fun mdist p q =
			sqrt (sq (real (ypos q - ypos p)) + 
			      sq (Point.dist p q))
		    fun meq p q = Point.eq p q andalso (ypos p = ypos q)
		in (mypos,mslide,mdist,meq)
		end
	in Point.pack x (S (methods y),ext)
	end;

(* new *)
fun new (x,y) = pack (x,y) %;
end;



(* line -------------------------------------------------------	*)

signature MINIPOINT =
sig
 
    type 'a subself
    type self = closed subself
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'a subself -> bool
    val dist : 'a subself -> 'a subself -> real
end;
 
signature LINE =
sig
    structure Param : MINIPOINT
    type ('a,'b) ext
    type ('a,'b) subself = ('a,'b) ext Root.subself
    type 'a self = ('a,closed) subself
    val unpack : ('a,'b) subself -> ('b -> ('a,'b) subself) * 'b
    val pack : ('a Param.subself * 'a Param.subself) ->
	       'b -> ('a,'b) subself
    val from : ('a,'b) subself -> 'a Param.subself
    val to : ('a,'b) subself -> 'a Param.subself
    val shift : ('a,'b) subself -> int -> ('a,'b) subself
    val length : ('a,'b) subself -> real
    val eq : ('a,'b) subself -> ('a,'b) subself -> bool
    val new : 'a Param.subself * 'a Param.subself -> 'a self
end;

functor Line (structure P : MINIPOINT) :> LINE where Param=P =
struct

structure Param = P;

(* types *)
datatype ('a,'b) state = S of
		(('a,'b) subself -> 'a Param.subself) *	      (* from   *)
		(('a,'b) subself -> 'a Param.subself) *	      (* to     *)
		(('a,'b) subself -> int -> ('a,'b) subself) * (* shift  *)
		(('a,'b) subself -> real) *                   (* length *)
		(('a,'b) subself -> ('a,'b) subself -> bool)  (* eq     *) 
withtype ('a,'b) ext = ('a,'b) state * 'b
and ('a,'b) subself = ('a,'b) ext Root.subself;
type 'a self = ('a,closed) subself;

(* messages *)
fun from l =    let val (_,(S (f,_,_,_,_),_)) = Root.unpack l in f l end;
fun to l =      let val (_,(S (_,t,_,_,_),_)) = Root.unpack l in t l end;
fun shift l n = let val (_,(S (_,_,s,_,_),_)) = Root.unpack l in s l n end;
fun length l =  let val (_,(S (_,_,_,e,_),_)) = Root.unpack l in e l end;
fun eq l l' =   let val (_,(S (_,_,_,_,e),_)) = Root.unpack l in e l l' end;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Root.unpack p);
fun pack (f,t) ext =
	let fun methods (f,t) =
		let fun mfr l = f
		    fun mto l = t
		    fun msh l n =
			repack (fn _ =>
				methods (Param.slide f n, Param.slide t n)) l
		    fun mlen l = Param.dist (from l) (to l)
		    fun meq p q =
			Param.eq (from p) (from q) andalso
			Param.eq (to p) (to q)
		in (mfr,mto,msh,mlen,meq)
		end
	in Root.pack () (S (methods (f,t)),ext)
	end;

(* new *)
fun new (f,t) = pack (f,t) %;
end;


structure PointLine = Line (structure P = Point);
structure CpointLine = Line (structure P = struct open Point; open Cpoint; end);
structure D2pointLine = Line (structure P = D2point);





