(* self-application, simple dispatch *)


use "../lib/smlprelude";


(* root objects -------------------------------------------------------	*)

open Closed;   (* closed datatype, % : closed *)

signature ROOT =
sig
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : unit -> 'a -> 'a subself
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a
type self = closed subself

(* packing *)
fun unpack ob = (I,ob);
fun pack () ob = ob;

(* new *)
fun new () = pack () %;
end;


(* eqobjs -------------------------------------------------------------	*)

signature EQOBJ =
sig
    type 'a ext
    type 'a subself = 'a ext Root.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> 'a subself -> bool) ->
		  ('a subself -> 'a subself -> bool)) ->
		 'a subself -> 'a subself
    val pack : unit -> 'a -> 'a subself
    exception Eq;
    val eq : 'a subself -> 'a subself -> bool
    val new : unit -> self
end;

structure Eqobj :> EQOBJ =
struct

(* types *)
datatype 'a state = S of
		('a subself -> 'a subself -> bool)	(* eq    *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Root.subself;
type self = closed subself;

(* messages *)
exception Eq;
fun eq p q = let val (_,(S e,_)) = Root.unpack p in e p q end;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Root.unpack p);
fun oeq (p : 'a subself) (q : 'a subself) : bool = raise Eq;
fun pack () ext = Root.pack () (S oeq,ext);

(* new *)
fun new () = pack () %;
end;


(* points -------------------------------------------------------------	*)

signature POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Eqobj.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> int) *
		  ('a subself -> int -> 'a subself) *
		  ('a subself -> 'a subself -> real) ->
		  ('a subself -> int) *
		  ('a subself -> int -> 'a subself) *
		  ('a subself -> 'a subself -> real)) ->
		 'a subself -> 'a subself
    val pack : int -> 'a -> 'a subself
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'a subself -> real
    val new : int -> self
end;

structure Point :> POINT =
struct

(* types *)
datatype 'a state = S of
		int *
		(('a subself -> int) *			(* pos	 *)
		 ('a subself -> int -> 'a subself) * 	(* slide *)
		 ('a subself -> 'a subself -> real))	(* dist  *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Eqobj.subself;
type self = closed subself;

(* messages *)
fun pos p     = let val (_,(S (_,(n,_,_)),_)) = Eqobj.unpack p in n p end;
fun slide p n = let val (_,(S (_,(_,s,_)),_)) = Eqobj.unpack p in s p n end;
fun dist p q  = let val (_,(S (_,(_,_,d)),_)) = Eqobj.unpack p in d p q end;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Eqobj.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s, m)) (Eqobj.unpack p);
fun mpos p = let val (_,(S (s,_),_)) = Eqobj.unpack p in s end;
fun mslide p n' = repacks (fn n => n + n') p;
fun mdist p q = real (pos q - pos p);
fun peq _ p q = pos p = pos q;
fun pack n ext = Eqobj.repack peq (Eqobj.pack () (S (n,(mpos,mslide,mdist)),ext));

(* new *)
fun new n = pack n %;
end;



(* colored points  ----------------------------------------------------	*)

signature CPOINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself;
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> color) *
		  ('a subself -> color -> 'a subself) ->
		  ('a subself -> color) *
		  ('a subself -> color -> 'a subself)) ->
		 'a subself -> 'a subself
    val pack : color * int -> 'a -> 'a subself
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
    val new : int * color -> self
end;

structure Cpoint :> CPOINT = 
struct

(* types *)
datatype 'a state = S of
		color *
		(('a subself -> color) *		(* col	 *)
		 ('a subself -> color -> 'a subself))	(* paint *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* messages *)
fun col p     = let val (_,(S (_,(c,_)),_)) = Point.unpack p in c p end;
fun paint p c = let val (_,(S (_,(_,t)),_)) = Point.unpack p in t p c end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Point.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Point.unpack p);
fun mcol p = let val (_,(S (s,_),_)) = Point.unpack p in s end;
fun mpaint p c' = repacks (fn c => c') p;
fun cpeq peq p q = peq p q andalso col p = col q;
fun pack (c,n) ext = Eqobj.repack cpeq (Point.pack n (S (c,(mcol,mpaint)),ext));

(* new *)
fun new (n,c) = pack (c,n) %;
end;



(* d2point ----------------------------------------------------	*)

signature D2POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> int) ->
		  ('a subself -> int)) ->
		 'a subself -> 'a subself
    val pack : int * int -> 'a -> 'a subself
    val xpos : 'a Point.subself -> int
    val ypos : 'a subself -> int
    val new : int * int -> self
end;

structure D2point :> D2POINT =
struct

(* types *)
datatype 'a state = S of
		int *
		('a subself -> int)		(* ypos  *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* messages *)
val xpos = Point.pos; (* alias *)
fun ypos p = let val (_,(S (_,y),_)) = Point.unpack p in y p end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Point.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Point.unpack p);
fun mypos p = let val (_,(S (s,_),_)) = Point.unpack p in s end;
fun d2pslide pslide p n = repacks (fn y => y+n) (pslide p n);
fun d2pdist pdist p q =	sqrt (sq (real (ypos q - ypos p)) + sq (pdist p q));
fun d2peq peq p q = peq p q andalso ypos p = ypos q;
fun pack (x,y) ext =
	Eqobj.repack d2peq
		(Point.repack (fn (mp,sp,dp) => (mp,d2pslide sp,d2pdist dp))
			(Point.pack x (S (y,mypos),ext)));

(* new *)
fun new (x,y) = pack (x,y) %;
end;



(* line -------------------------------------------------------	*)

signature LINE =
sig
    type ('a,'b) ext
    type ('a,'b) subself = ('a,'b) ext Eqobj.subself
    type 'a self = ('a,closed) subself
    val unpack : ('a,'b) subself -> ('b -> ('a,'b) subself) * 'b
    val repack : ((('a,'b) subself -> 'a Point.subself) *
		  (('a,'b) subself -> 'a Point.subself) *
		  (('a,'b) subself -> int -> ('a,'b) subself) *
		  (('a,'b) subself -> real) ->
		  (('a,'b) subself -> 'a Point.subself) *
		  (('a,'b) subself -> 'a Point.subself) *
		  (('a,'b) subself -> int -> ('a,'b) subself) *
		  (('a,'b) subself -> real)) ->
		 ('a,'b) subself -> ('a,'b) subself
    val pack : ('a Point.subself * 'a Point.subself) -> 'b -> ('a,'b) subself
    val from : ('a,'b) subself -> 'a Point.subself
    val to : ('a,'b) subself -> 'a Point.subself
    val shift : ('a,'b) subself -> int -> ('a,'b) subself
    val length : ('a,'b) subself -> real
    val new : 'a Point.subself * 'a Point.subself -> 'a self
end;

structure Line :> LINE =
struct

(* as a pair of homogeneous points *)
datatype ('a,'b) state = S of
		('a Point.subself * 'a Point.subself) *
		((('a,'b) subself -> 'a Point.subself) *	(* from *)
		 (('a,'b) subself -> 'a Point.subself) *	(* to   *)
		 (('a,'b) subself -> int -> ('a,'b) subself) *  (* shift *)
		 (('a,'b) subself -> real))                     (* length *)
withtype ('a,'b) ext = ('a,'b) state * 'b
and ('a,'b) subself = ('a,'b) ext Eqobj.subself;
type 'a self = ('a,closed) subself;

(* messages *)
fun from l    = let val (_,(S (_,(f,_,_,_)),_)) = Eqobj.unpack l in f l end;
fun to l      = let val (_,(S (_,(_,t,_,_)),_)) = Eqobj.unpack l in t l end;
fun shift l n = let val (_,(S (_,(_,_,s,_)),_)) = Eqobj.unpack l in s l n end;
fun length l  = let val (_,(S (_,(_,_,_,e)),_)) = Eqobj.unpack l in e l end;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Eqobj.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Eqobj.unpack p);
fun state p = let val (_,(S (s,_),_)) = Eqobj.unpack p in s end;
fun state p = let val (_,(S (s,_),_)) = Eqobj.unpack p in s end;
fun mfr l = let val (f,_) = state l in f end;
fun mto l = let val (_,t) = state l in t end;
fun msh l n = repacks (fn (f,t) => (Point.slide f n, Point.slide t n)) l;
fun mlen l = Point.dist (from l) (to l);
fun leq _ p q =	Eqobj.eq (from p) (from q) andalso Eqobj.eq (to p) (to q);
fun pack (f,t) ext =
	Eqobj.repack leq (Eqobj.pack () (S ((f,t),(mfr,mto,msh,mlen)),ext))

(* new *)
fun new (f,t) = pack (f,t) %;
end;



